from reconfigure.configs.base import Reconfig
from reconfigure.parsers import BIND9Parser
from reconfigure.includers import BIND9Includer
from reconfigure.builders import BoundBuilder
from reconfigure.items.bind9 import BIND9Data


class BIND9Config (Reconfig):
    """
    ``named.conf``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': BIND9Parser(),
            'includer': BIND9Includer(),
            'builder': BoundBuilder(BIND9Data),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData
from reconfigure.items.util import yn_getter, yn_setter


class NetatalkData (BoundData):
    pass


class GlobalData (BoundData):
    pass


class ShareData (BoundData):
    fields = ['path', 'appledouble', 'valid users', 'cnid scheme', 'ea', 'password']
    defaults = ['', 'ea', '', 'dbd', 'none', '']

    def template(self):
        return Node(
            'share',
            *[PropertyNode(x, y) for x, y in zip(ShareData.fields, ShareData.defaults)]
        )


NetatalkData.bind_child('global', lambda x: x.get('Global'), item_class=GlobalData)
NetatalkData.bind_collection('shares', selector=lambda x: x.name != 'Global', item_class=ShareData)


GlobalData.bind_property('afp port', 'afp_port', default='548')
GlobalData.bind_property('cnid listen', 'cnid_listen', default='localhost:4700')
GlobalData.bind_property('uam list', 'uam_list', default='uams_dhx.so,uams_dhx2.so')
GlobalData.bind_property(
    'zeroconf', 'zeroconf', default=True,
    getter=yn_getter, setter=yn_setter)

ShareData.bind_name('name')
ShareData.bind_attribute('comment', 'comment', path=lambda x: x.get('path'), default='')
for f, d in zip(ShareData.fields, ShareData.defaults):
    ShareData.bind_property(f, f.replace(' ', '_'), default=d)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class BIND9Data (BoundData):
    pass


class ZoneData (BoundData):
    def template(self):
        return Node(
            'zone',
            PropertyNode('type', 'master'),
            PropertyNode('file', 'db.example.com'),
            parameter='"example.com"',
        )


quote = lambda x: '"%s"' % x
unquote = lambda x: x.strip('"')

BIND9Data.bind_collection('zones', selector=lambda x: x.name == 'zone', item_class=ZoneData)
ZoneData.bind_attribute('parameter', 'name', getter=unquote, setter=quote)
ZoneData.bind_property('type', 'type')
ZoneData.bind_property('file', 'file', getter=unquote, setter=quote)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData
from reconfigure.items.util import yn_getter, yn_setter


class SambaData (BoundData):
    pass


class GlobalData (BoundData):
    pass


class ShareData (BoundData):
    fields = [
        'comment', 'path', 'guest ok', 'browseable', 'create mask', 'directory mask', 'read only',
        'follow symlinks', 'wide links', 'fstype', 'write list', 'veto files',
        'force create mode', 'force directory mode', 'dfree command', 'force user', 'force group',
        'valid_users', 'read_list',
    ]
    defaults = [
        '', '', 'no', 'yes', '0744', '0755', 'yes',
        'yes', 'no', 'NTFS', '', '', '000', '000', '',
        '', '', '', '',
    ]
    default_values = [
        '', '', False, True, '0744', '0755', True,
        True, False, '', '', '', '000', '000', '',
        '', '', '', '',
    ]

    def template(self):
        return Node(
            'share',
            *[PropertyNode(x, y) for x, y in zip(ShareData.fields, ShareData.defaults)]
        )


SambaData.bind_child('global', lambda x: x.get('global'), item_class=GlobalData)
SambaData.bind_collection('shares', selector=lambda x: x.name != 'global', item_class=ShareData)


GlobalData.bind_property('workgroup', 'workgroup', default='')
GlobalData.bind_property('server string', 'server_string', default='')
GlobalData.bind_property('interfaces', 'interfaces', default='')
GlobalData.bind_property(
    'bind interfaces only', 'bind_interfaces_only', default=True,
    getter=yn_getter, setter=yn_setter)
GlobalData.bind_property('log file', 'log_file', default='')
GlobalData.bind_property('security', 'security', default='user')

ShareData.bind_name('name')
for f, d in zip(ShareData.fields, ShareData.default_values):
    if d not in [True, False]:
        ShareData.bind_property(f, f.replace(' ', '_'), default=d)
    else:
        ShareData.bind_property(
            f, f.replace(' ', '_'), default=d,
            getter=yn_getter, setter=yn_setter)



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser


class SSVParser (BaseParser):
    """
    A parser for files containing space-separated value (notably, ``/etc/fstab`` and friends)

    :param separator: separator character, defaults to whitespace
    :param maxsplit: max number of tokens per line, defaults to infinity
    :param comment: character denoting comments
    :param continuation: line continuation character, None to disable
    """

    def __init__(self, separator=None, maxsplit=-1, comment='#', continuation=None, *args, **kwargs):
        self.separator = separator
        self.maxsplit = maxsplit
        self.comment = comment
        self.continuation = continuation
        BaseParser.__init__(self, *args, **kwargs)

    def parse(self, content):
        rawlines = content.splitlines()
        lines = []
        while rawlines:
            l = rawlines.pop(0).strip()
            while self.continuation and rawlines and l.endswith(self.continuation):
                l = l[:-len(self.continuation)]
                l += rawlines.pop(0)
            lines.append(l)
        root = RootNode()
        last_comment = None
        for line in lines:
            line = line.strip()
            if line:
                if line.startswith(self.comment):
                    c = line.strip(self.comment).strip()
                    if last_comment:
                        last_comment += '\n' + c
                    else:
                        last_comment = c
                    continue
                if len(line) == 0:
                    continue
                tokens = line.split(self.separator, self.maxsplit)
                node = Node('line')
                if last_comment:
                    node.comment = last_comment
                    last_comment = None
                for token in tokens:
                    if token.startswith(self.comment):
                        node.comment = ' '.join(tokens[tokens.index(token):])[1:].strip()
                        break
                    node.append(Node(
                        name='token',
                        children=[
                            PropertyNode(name='value', value=token)
                        ]
                    ))
                root.append(node)
        return root

    def stringify(self, tree):
        r = ''
        for node in tree.children:
            if node.comment and '\n' in node.comment:
                r += ''.join('%s %s\n' % (self.comment, x) for x in node.comment.splitlines())
            r += (self.separator or '\t').join(x.get('value').value for x in node.children)
            if node.comment and not '\n' in node.comment:
                r += ' # %s' % node.comment
            r += '\n'
        return r



from reconfigure.nodes import RootNode, Node, PropertyNode
from reconfigure.parsers import BaseParser


class CrontabParser(BaseParser):

    def __init__(self, remove_comments=False):
        self.remove_comments = remove_comments

    def parse(self, content):
        root = RootNode()
        lines = [l.strip() for l in content.splitlines() if l]
        comment = None
        for line in lines:
            if line.startswith('#'):
                comment = '\n'.join([comment, line]) if comment else line[1:]
                continue
            elif line.startswith('@'):
                special, command = line.split(' ', 1)
                node = Node('special_task', comment=comment)
                node.append(PropertyNode('special', special))
                node.append(PropertyNode('command', command))

            else:
                split_line = line.split(' ', 5)
                if len(split_line) <= 3 and '=' in line:
                    name, value = [n.strip() for n in line.split('=')]
                    if not name:
                        continue
                    node = Node('env_setting', comment=comment)
                    node.append(PropertyNode('name', name))
                    node.append(PropertyNode('value', value))
                elif len(split_line) == 6:
                    node = Node('normal_task', comment=comment)
                    node.append(PropertyNode('minute', split_line[0]))
                    node.append(PropertyNode('hour', split_line[1]))
                    node.append(PropertyNode('day_of_month', split_line[2]))
                    node.append(PropertyNode('month', split_line[3]))
                    node.append(PropertyNode('day_of_week', split_line[4]))
                    node.append(PropertyNode('command', split_line[5]))
                else:
                    continue
            root.append(node)
            comment = None
        root.comment = comment
        return root

    def stringify(self, tree):
        result_lines = []
        stringify_func = {
            'special_task': self.stringify_special_task,
            'env_setting': self.stringify_env_setting,
            'normal_task': self.stringify_normal_task,
        }
        for node in tree:
            if isinstance(node, Node):
                string_line = stringify_func.get(node.name, lambda x: '')(node)
                if node.comment:
                    result_lines.append('#' + node.comment)
                result_lines.append(string_line)
        return '\n'.join([line for line in result_lines if line])

    def stringify_special_task(self, node):
        special_node = node.get('special')
        command_node = node.get('command')
        if isinstance(special_node, PropertyNode) and isinstance(command_node, PropertyNode):
            return ' '.join([special_node.value, command_node.value])
        return ''

    def stringify_env_setting(self, node):
        name = node.get('name')
        value = node.get('value')
        if isinstance(name, PropertyNode) and isinstance(value, PropertyNode):
                return ' = '.join([name.value, value.value])
        return ''

    def stringify_normal_task(self, node):
        if all([isinstance(child, PropertyNode) for child in node.children]):
            values_list = [str(pr_node.value).strip() for pr_node in node.children if pr_node.value]
            if len(values_list) == 6:
                return ' '.join(values_list)
        return ''



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData
from reconfigure.items.util import onezero_getter, onezero_setter


class CSFData (BoundData):
    pass


CSFData.bind_property('TESTING', 'testing', getter=onezero_getter, setter=onezero_setter)
CSFData.bind_property('IPV6', 'ipv6', getter=onezero_getter, setter=onezero_setter)

CSFData.bind_property('TCP_IN', 'tcp_in')
CSFData.bind_property('TCP_OUT', 'tcp_out')
CSFData.bind_property('UDP_IN', 'udp_in')
CSFData.bind_property('UDP_OUT', 'udp_out')
CSFData.bind_property('TCP6_IN', 'tcp6_in')
CSFData.bind_property('TCP6_OUT', 'tcp6_out')
CSFData.bind_property('UDP6_IN', 'udp6_in')
CSFData.bind_property('UDP6_OUT', 'udp6_out')
CSFData.bind_property('ETH_DEVICE', 'eth_device')
CSFData.bind_property('ETH6_DEVICE', 'eth6_device')



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import ShellParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.csf import CSFData


class CSFConfig (Reconfig):
    """
    ``CSF main config``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': ShellParser(),
            'builder': BoundBuilder(CSFData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.nodes import *
from reconfigure.parsers.nginx import NginxParser


class BIND9Parser (NginxParser):
    """
    A parser for named.conf
    """

    tokens = [
        (r"[\w_]+\s*?.*?{", lambda s, t: ('section_start', t)),
        (r"[\w\d_:.]+?.*?;", lambda s, t: ('option', t)),
        (r"\".*?\"\s*;", lambda s, t: ('option', t)),
        (r"\s", lambda s, t: 'whitespace'),
        (r"$^", lambda s, t: 'newline'),
        (r"\#.*?\n", lambda s, t: ('comment', t)),
        (r"//.*?\n", lambda s, t: ('comment', t)),
        (r"/\*.*?\*/", lambda s, t: ('comment', t)),
        (r"\};", lambda s, t: 'section_end'),
    ]
    token_section_end = '};'



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import CrontabParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.crontab import CrontabData


class CrontabConfig (Reconfig):
    def __init__(self, **kwargs):
        k = {
            'parser': CrontabParser(),
            'builder': BoundBuilder(CrontabData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import IniFileParser
from reconfigure.includers import SupervisorIncluder
from reconfigure.builders import BoundBuilder
from reconfigure.items.supervisor import SupervisorData


class SupervisorConfig (Reconfig):
    """
    ``/etc/supervisor/supervisord.conf``
    """

    def __init__(self, **kwargs):
        k = {
            'parser': IniFileParser(),
            'includer': SupervisorIncluder(),
            'builder': BoundBuilder(SupervisorData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData
from reconfigure.items.util import yn_getter, yn_setter


class CTDBData (BoundData):
    pass

CTDBData.bind_property('CTDB_RECOVERY_LOCK', 'recovery_lock_file', path=lambda x: x.get(None))
CTDBData.bind_property('CTDB_PUBLIC_INTERFACE', 'public_interface', path=lambda x: x.get(None))
CTDBData.bind_property('CTDB_PUBLIC_ADDRESSES', 'public_addresses_file', default='/etc/ctdb/public_addresses', path=lambda x: x.get(None))
CTDBData.bind_property(
    'CTDB_MANAGES_SAMBA', 'manages_samba', path=lambda x: x.get(None),
    getter=yn_getter, setter=yn_setter)
CTDBData.bind_property('CTDB_NODES', 'nodes_file', default='/etc/ctdb/nodes', path=lambda x: x.get(None))
CTDBData.bind_property('CTDB_LOGFILE', 'log_file', path=lambda x: x.get(None))
CTDBData.bind_property('CTDB_DEBUGLEVEL', 'debug_level', default='2', path=lambda x: x.get(None))
CTDBData.bind_property('CTDB_PUBLIC_NETWORK', 'public_network', default='', path=lambda x: x.get(None))
CTDBData.bind_property('CTDB_PUBLIC_GATEWAY', 'public_gateway', default='', path=lambda x: x.get(None))


class NodesData (BoundData):
    pass


class NodeData (BoundData):
    def template(self):
        return Node('line', children=[
            Node('token', children=[PropertyNode('value', '127.0.0.1')]),
        ])


NodesData.bind_collection('nodes', item_class=NodeData)
NodeData.bind_property('value', 'address', path=lambda x: x.children[0])


class PublicAddressesData (BoundData):
    pass


class PublicAddressData (BoundData):
    def template(self):
        return Node('line', children=[
            Node('token', children=[PropertyNode('value', '127.0.0.1')]),
            Node('token', children=[PropertyNode('value', 'eth0')]),
        ])

PublicAddressesData.bind_collection('addresses', item_class=PublicAddressData)
PublicAddressData.bind_property('value', 'address', path=lambda x: x.children[0])
PublicAddressData.bind_property('value', 'interface', path=lambda x: x.children[1])



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class CrontabData(BoundData):
    """Data class for crontab configs"""
    pass


class CrontabNormalTaskData(BoundData):
    fields = ['minute', 'hour', 'day_of_month', 'month', 'day_of_week', 'command']

    def describe(self):
        return ' '.join(getattr(self, x) for x in self.fields)

    def template(self, **kwargs):
        return Node('normal_task', children=[
            PropertyNode('minute', '0'),
            PropertyNode('hour', '0'),
            PropertyNode('day_of_month', '1'),
            PropertyNode('month', '1'),
            PropertyNode('day_of_week', '1'),
            PropertyNode('command', 'false')
        ])


class CrontabSpecialTaskData(BoundData):
    fields = ['special', 'command']

    def template(self, **kwargs):
        return Node('special_task', children=[
            PropertyNode('special', '@reboot'),
            PropertyNode('command', 'false')
        ])


class CrontabEnvSettingData(BoundData):
    fields = ['name', 'value']

    def template(self, **kwargs):
        return Node('env_setting', children=[
            PropertyNode('name', 'ENV_NAME'),
            PropertyNode('value', 'ENV_VALUE')
        ])


def bind_for_fields(bound_data_class):
    for field in bound_data_class.fields:
        bound_data_class.bind_property(field, field)

CrontabData.bind_collection('normal_tasks', selector=lambda x: x.name == 'normal_task', item_class=CrontabNormalTaskData)
bind_for_fields(CrontabNormalTaskData)

CrontabNormalTaskData.bind_attribute('comment', 'comment')

CrontabData.bind_collection('env_settings', selector=lambda x: x.name == 'env_setting', item_class=CrontabEnvSettingData)
bind_for_fields(CrontabEnvSettingData)

CrontabEnvSettingData.bind_attribute('comment', 'comment')

CrontabData.bind_collection('special_tasks', selector=lambda x: x.name == 'special_task', item_class=CrontabSpecialTaskData)
bind_for_fields(CrontabSpecialTaskData)

CrontabSpecialTaskData.bind_attribute('comment', 'comment')



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import IniFileParser
from reconfigure.parsers import SSVParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.ctdb import CTDBData, NodesData, PublicAddressesData


class CTDBConfig (Reconfig):
    """
    ``CTDB main config``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': IniFileParser(sectionless=True),
            'builder': BoundBuilder(CTDBData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)


class CTDBNodesConfig (Reconfig):
    """
    ``CTDB node list file``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': SSVParser(),
            'builder': BoundBuilder(NodesData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)


class CTDBPublicAddressesConfig (Reconfig):
    """
    ``CTDB public address list file``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': SSVParser(separator=' '),
            'builder': BoundBuilder(PublicAddressesData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)
